pub const SNAPSHOT_KEY: &str = "snapshot";
pub const LAST_INDEX_KEY: &str = "last_index";
pub const HARD_STATE_KEY: &str = "hard_state";
pub const CONF_STATE_KEY: &str = "conf_state";

pub const ENTRY_KEY_LENGTH: usize = 10;
